import unittest
from flowhigh.format.formatting import _process_data


class FormatterTest(unittest.TestCase):

    @classmethod
    def setup(cls, marked_sql: str, option: str = "compact") -> str:
        return _process_data(marked_sql, option)

    def test_second_new_line(self):
        marked_sql = """SELECT p1 ✖,p2 ✖"""

        expected = """SELECT p1 
,p2 """
        self.assertEqual(expected, self.setup(marked_sql))

    def test_anchor(self):
        marked_sql = """SELECT⚓ p1 ✖⚓,p2 ✖⚓"""

        expected = """SELECT p1 
      ,p2 
      """

        self.assertEqual(expected, self.setup(marked_sql))

    def test_nested_anchor(self):
        marked_sql = """❌FROM⚓(➕SELECT⚓ op1.OrderID ✖⚓,op1.ProductID as ⛓⚓p1"""

        expected = """  FROM(SELECT op1.OrderID 
             ,op1.ProductID as p1"""

        self.assertEqual(expected, self.setup(marked_sql))

    def test_optional_new_line(self):
        marked_sql = """❌FROM⚓(➕SELECT⚓ op1.OrderID ✖⚓,op1.ProductID as ⛓⚓p1"""

        expected = """  FROM(SELECT op1.OrderID 
             ,op1.ProductID as 
             p1"""

        self.assertEqual(expected, self.setup(marked_sql, "comfortable"))

    def test_compact_marker(self):
        marked_sql = """op1.OrderID⚫=⚫op2.OrderID"""

        expected = """op1.OrderID=op2.OrderID"""

        self.assertEqual(expected, self.setup(marked_sql))

    def test_first_new_line(self):
        marked_sql = """SELECT ❌FROM"""

        expected = """SELECT 
FROM"""

        self.assertEqual(expected, self.setup(marked_sql))

    def test_first_new_line(self):
        marked_sql = """SELECT ❌FROM"""

        expected = """SELECT 
FROM"""

        self.assertEqual(expected, self.setup(marked_sql))

    def test_sample_regular_compact(self):
        marked_sql = """WITH← all_sales AS(➕❌SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,SUM(➕sales_cnt➖) AS ⛓⚓sales_cnt ✖⚓,SUM(➕sales_amt➖) AS ⛓⚓sales_amt❌FROM⚓(➕SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,cs.quantity ⚫-⚫ COALESCE(➕cr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,cs.ext_sales_price ⚫-⚫ COALESCE(➕cr.return_amount✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ catalog_sales cs ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ cs.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ cs.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ catalog_returns cr ❌ON⚓(➕cs.order_number ⚫=⚫ cr.order_number ✖AND⚓ cs.item_sk ⚫=⚫ cr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ss.quantity ⚫-⚫ COALESCE(➕sr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ss.ext_sales_price ⚫-⚫ COALESCE(➕sr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ store_sales ss ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ss.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ss.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ store_returns sr ❌ON⚓(➕ss.ticket_number ⚫=⚫ sr.ticket_number ✖AND⚓ ss.item_sk ⚫=⚫ sr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ws.quantity ⚫-⚫ COALESCE(➕wr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ws.ext_sales_price ⚫-⚫ COALESCE(➕wr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ web_sales ws ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ws.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ws.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ web_returns wr ❌ON⚓(➕ws.order_number ⚫=⚫ wr.order_number ✖AND⚓ ws.item_sk ⚫=⚫ wr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes'➖) sales_detail ❌GROUP⚓ BY d.year✖⚓,i.brand_id✖⚓,i.class_id✖⚓,i.category_id✖⚓,i.manufact_id⛓⚓➖)❌SELECT⚓ prev_yr.d_year AS ⛓⚓prev_year ✖⚓,curr_yr.d_year AS ⛓⚓year ✖⚓,curr_yr.i_brand_id ✖⚓,curr_yr.i_class_id ✖⚓,curr_yr.i_category_id ✖⚓,curr_yr.i_manufact_id ✖⚓,prev_yr.sales_cnt AS ⛓⚓prev_yr_cnt ✖⚓,curr_yr.sales_cnt AS ⛓⚓curr_yr_cnt ✖⚓,curr_yr.sales_cnt ⚫-⚫ prev_yr.sales_cnt AS ⛓⚓sales_cnt_diff ✖⚓,curr_yr.sales_amt ⚫-⚫ prev_yr.sales_amt AS ⛓⚓sales_amt_diff❌FROM⚓ all_sales curr_yr✖⚓,all_sales prev_yr❌WHERE⚓ curr_yr.brand_id ⚫=⚫ prev_yr.brand_id ✖AND⚓ curr_yr.class_id ⚫=⚫ prev_yr.class_id ✖AND⚓ curr_yr.category_id ⚫=⚫ prev_yr.category_id ✖AND⚓ curr_yr.manufact_id ⚫=⚫ prev_yr.manufact_id ✖AND⚓ curr_yr.year ⚫=⚫ 2000 ✖AND⚓ prev_yr.year ⚫=⚫ 2000 ⚫-⚫ 1 ✖AND⚓ CAST(➕curr_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫/⚫ CAST(➕prev_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫<⚫ 0.9❌ORDER⚓ BY sales_cnt_diff❌limit⚓ 100"""

        compact = """WITH all_sales AS(
    SELECT d.year 
          ,i.brand_id 
          ,i.class_id 
          ,i.category_id 
          ,i.manufact_id 
          ,SUM(sales_cnt) AS sales_cnt 
          ,SUM(sales_amt) AS sales_amt
      FROM(SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,cs.quantity-COALESCE(cr.return_quantity
                  ,0) AS sales_cnt 
                 ,cs.ext_sales_price-COALESCE(cr.return_amount
                  ,0.0) AS sales_amt 
             FROM catalog_sales cs 
             JOIN item i 
               ON i.item_sk=cs.item_sk
             JOIN date_dim d 
               ON d.date_sk=cs.sold_date_sk
             LEFT 
             JOIN catalog_returns cr 
               ON(cs.order_number=cr.order_number
               AND cs.item_sk=cr.item_sk)
            WHERE i.category='Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ss.quantity-COALESCE(sr.return_quantity
                  ,0) AS sales_cnt 
                 ,ss.ext_sales_price-COALESCE(sr.return_amt
                  ,0.0) AS sales_amt 
             FROM store_sales ss 
             JOIN item i 
               ON i.item_sk=ss.item_sk
             JOIN date_dim d 
               ON d.date_sk=ss.sold_date_sk
             LEFT 
             JOIN store_returns sr 
               ON(ss.ticket_number=sr.ticket_number
               AND ss.item_sk=sr.item_sk)
            WHERE i.category='Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ws.quantity-COALESCE(wr.return_quantity
                  ,0) AS sales_cnt 
                 ,ws.ext_sales_price-COALESCE(wr.return_amt
                  ,0.0) AS sales_amt 
             FROM web_sales ws 
             JOIN item i 
               ON i.item_sk=ws.item_sk
             JOIN date_dim d 
               ON d.date_sk=ws.sold_date_sk
             LEFT 
             JOIN web_returns wr 
               ON(ws.order_number=wr.order_number
               AND ws.item_sk=wr.item_sk)
            WHERE i.category='Shoes') sales_detail
     GROUP BY d.year
          ,i.brand_id
          ,i.class_id
          ,i.category_id
          ,i.manufact_id)
SELECT prev_yr.d_year AS prev_year 
      ,curr_yr.d_year AS year 
      ,curr_yr.i_brand_id 
      ,curr_yr.i_class_id 
      ,curr_yr.i_category_id 
      ,curr_yr.i_manufact_id 
      ,prev_yr.sales_cnt AS prev_yr_cnt 
      ,curr_yr.sales_cnt AS curr_yr_cnt 
      ,curr_yr.sales_cnt-prev_yr.sales_cnt AS sales_cnt_diff
      ,curr_yr.sales_amt-prev_yr.sales_amt AS sales_amt_diff
  FROM all_sales curr_yr
      ,all_sales prev_yr
 WHERE curr_yr.brand_id=prev_yr.brand_id
   AND curr_yr.class_id=prev_yr.class_id
   AND curr_yr.category_id=prev_yr.category_id
   AND curr_yr.manufact_id=prev_yr.manufact_id
   AND curr_yr.year=2000
   AND prev_yr.year=2000-1
   AND CAST(curr_yr.sales_cnt AS DECIMAL(17
                 ,2))/CAST(prev_yr.sales_cnt AS DECIMAL(17
                 ,2))<0.9
 ORDER BY sales_cnt_diff
 limit 100"""

        self.assertEqual(compact, _process_data(marked_sql, "compact"))

    def test_sample_regular_balanced(self):
        marked_sql = """WITH← all_sales AS(➕❌SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,SUM(➕sales_cnt➖) AS ⛓⚓sales_cnt ✖⚓,SUM(➕sales_amt➖) AS ⛓⚓sales_amt❌FROM⚓(➕SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,cs.quantity ⚫-⚫ COALESCE(➕cr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,cs.ext_sales_price ⚫-⚫ COALESCE(➕cr.return_amount✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ catalog_sales cs ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ cs.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ cs.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ catalog_returns cr ❌ON⚓(➕cs.order_number ⚫=⚫ cr.order_number ✖AND⚓ cs.item_sk ⚫=⚫ cr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ss.quantity ⚫-⚫ COALESCE(➕sr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ss.ext_sales_price ⚫-⚫ COALESCE(➕sr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ store_sales ss ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ss.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ss.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ store_returns sr ❌ON⚓(➕ss.ticket_number ⚫=⚫ sr.ticket_number ✖AND⚓ ss.item_sk ⚫=⚫ sr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ws.quantity ⚫-⚫ COALESCE(➕wr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ws.ext_sales_price ⚫-⚫ COALESCE(➕wr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ web_sales ws ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ws.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ws.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ web_returns wr ❌ON⚓(➕ws.order_number ⚫=⚫ wr.order_number ✖AND⚓ ws.item_sk ⚫=⚫ wr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes'➖) sales_detail ❌GROUP⚓ BY d.year✖⚓,i.brand_id✖⚓,i.class_id✖⚓,i.category_id✖⚓,i.manufact_id⛓⚓➖)❌SELECT⚓ prev_yr.d_year AS ⛓⚓prev_year ✖⚓,curr_yr.d_year AS ⛓⚓year ✖⚓,curr_yr.i_brand_id ✖⚓,curr_yr.i_class_id ✖⚓,curr_yr.i_category_id ✖⚓,curr_yr.i_manufact_id ✖⚓,prev_yr.sales_cnt AS ⛓⚓prev_yr_cnt ✖⚓,curr_yr.sales_cnt AS ⛓⚓curr_yr_cnt ✖⚓,curr_yr.sales_cnt ⚫-⚫ prev_yr.sales_cnt AS ⛓⚓sales_cnt_diff ✖⚓,curr_yr.sales_amt ⚫-⚫ prev_yr.sales_amt AS ⛓⚓sales_amt_diff❌FROM⚓ all_sales curr_yr✖⚓,all_sales prev_yr❌WHERE⚓ curr_yr.brand_id ⚫=⚫ prev_yr.brand_id ✖AND⚓ curr_yr.class_id ⚫=⚫ prev_yr.class_id ✖AND⚓ curr_yr.category_id ⚫=⚫ prev_yr.category_id ✖AND⚓ curr_yr.manufact_id ⚫=⚫ prev_yr.manufact_id ✖AND⚓ curr_yr.year ⚫=⚫ 2000 ✖AND⚓ prev_yr.year ⚫=⚫ 2000 ⚫-⚫ 1 ✖AND⚓ CAST(➕curr_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫/⚫ CAST(➕prev_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫<⚫ 0.9❌ORDER⚓ BY sales_cnt_diff❌limit⚓ 100"""

        expected = """WITH all_sales AS(
    SELECT d.year 
          ,i.brand_id 
          ,i.class_id 
          ,i.category_id 
          ,i.manufact_id 
          ,SUM(sales_cnt) AS sales_cnt 
          ,SUM(sales_amt) AS sales_amt
      FROM(SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,cs.quantity - COALESCE(cr.return_quantity
                  ,0) AS sales_cnt 
                 ,cs.ext_sales_price - COALESCE(cr.return_amount
                  ,0.0) AS sales_amt 
             FROM catalog_sales cs 
             JOIN item i 
               ON i.item_sk = cs.item_sk
             JOIN date_dim d 
               ON d.date_sk = cs.sold_date_sk
             LEFT 
             JOIN catalog_returns cr 
               ON(cs.order_number = cr.order_number
               AND cs.item_sk = cr.item_sk)
            WHERE i.category = 'Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ss.quantity - COALESCE(sr.return_quantity
                  ,0) AS sales_cnt 
                 ,ss.ext_sales_price - COALESCE(sr.return_amt
                  ,0.0) AS sales_amt 
             FROM store_sales ss 
             JOIN item i 
               ON i.item_sk = ss.item_sk
             JOIN date_dim d 
               ON d.date_sk = ss.sold_date_sk
             LEFT 
             JOIN store_returns sr 
               ON(ss.ticket_number = sr.ticket_number
               AND ss.item_sk = sr.item_sk)
            WHERE i.category = 'Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ws.quantity - COALESCE(wr.return_quantity
                  ,0) AS sales_cnt 
                 ,ws.ext_sales_price - COALESCE(wr.return_amt
                  ,0.0) AS sales_amt 
             FROM web_sales ws 
             JOIN item i 
               ON i.item_sk = ws.item_sk
             JOIN date_dim d 
               ON d.date_sk = ws.sold_date_sk
             LEFT 
             JOIN web_returns wr 
               ON(ws.order_number = wr.order_number
               AND ws.item_sk = wr.item_sk)
            WHERE i.category = 'Shoes') sales_detail
     GROUP BY d.year
          ,i.brand_id
          ,i.class_id
          ,i.category_id
          ,i.manufact_id)
SELECT prev_yr.d_year AS prev_year 
      ,curr_yr.d_year AS year 
      ,curr_yr.i_brand_id 
      ,curr_yr.i_class_id 
      ,curr_yr.i_category_id 
      ,curr_yr.i_manufact_id 
      ,prev_yr.sales_cnt AS prev_yr_cnt 
      ,curr_yr.sales_cnt AS curr_yr_cnt 
      ,curr_yr.sales_cnt - prev_yr.sales_cnt AS sales_cnt_diff
      ,curr_yr.sales_amt - prev_yr.sales_amt AS sales_amt_diff
  FROM all_sales curr_yr
      ,all_sales prev_yr
 WHERE curr_yr.brand_id = prev_yr.brand_id
   AND curr_yr.class_id = prev_yr.class_id
   AND curr_yr.category_id = prev_yr.category_id
   AND curr_yr.manufact_id = prev_yr.manufact_id
   AND curr_yr.year = 2000
   AND prev_yr.year = 2000 - 1
   AND CAST(curr_yr.sales_cnt AS DECIMAL(17
                 ,2)) / CAST(prev_yr.sales_cnt AS DECIMAL(17
                 ,2)) < 0.9
 ORDER BY sales_cnt_diff
 limit 100"""

        self.assertEqual(expected, _process_data(marked_sql, "balanced"))

    def test_sample_regular_comfortable(self):
        marked_sql = """WITH← all_sales AS(➕❌SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,SUM(➕sales_cnt➖) AS ⛓⚓sales_cnt ✖⚓,SUM(➕sales_amt➖) AS ⛓⚓sales_amt❌FROM⚓(➕SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,cs.quantity ⚫-⚫ COALESCE(➕cr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,cs.ext_sales_price ⚫-⚫ COALESCE(➕cr.return_amount✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ catalog_sales cs ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ cs.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ cs.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ catalog_returns cr ❌ON⚓(➕cs.order_number ⚫=⚫ cr.order_number ✖AND⚓ cs.item_sk ⚫=⚫ cr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ss.quantity ⚫-⚫ COALESCE(➕sr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ss.ext_sales_price ⚫-⚫ COALESCE(➕sr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ store_sales ss ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ss.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ss.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ store_returns sr ❌ON⚓(➕ss.ticket_number ⚫=⚫ sr.ticket_number ✖AND⚓ ss.item_sk ⚫=⚫ sr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes' ❌UNION⚓❌ SELECT⚓ d.year ✖⚓,i.brand_id ✖⚓,i.class_id ✖⚓,i.category_id ✖⚓,i.manufact_id ✖⚓,ws.quantity ⚫-⚫ COALESCE(➕wr.return_quantity✖⚓,0➖) AS ⛓⚓sales_cnt ✖⚓,ws.ext_sales_price ⚫-⚫ COALESCE(➕wr.return_amt✖⚓,0.0➖) AS ⛓⚓sales_amt ❌FROM⚓ web_sales ws ❌JOIN⚓ item i ❌ON⚓ i.item_sk ⚫=⚫ ws.item_sk ❌JOIN⚓ date_dim d ❌ON⚓ d.date_sk ⚫=⚫ ws.sold_date_sk ❌LEFT⚓ ❌JOIN⚓ web_returns wr ❌ON⚓(➕ws.order_number ⚫=⚫ wr.order_number ✖AND⚓ ws.item_sk ⚫=⚫ wr.item_sk➖) ❌WHERE⚓ i.category ⚫=⚫ 'Shoes'➖) sales_detail ❌GROUP⚓ BY d.year✖⚓,i.brand_id✖⚓,i.class_id✖⚓,i.category_id✖⚓,i.manufact_id⛓⚓➖)❌SELECT⚓ prev_yr.d_year AS ⛓⚓prev_year ✖⚓,curr_yr.d_year AS ⛓⚓year ✖⚓,curr_yr.i_brand_id ✖⚓,curr_yr.i_class_id ✖⚓,curr_yr.i_category_id ✖⚓,curr_yr.i_manufact_id ✖⚓,prev_yr.sales_cnt AS ⛓⚓prev_yr_cnt ✖⚓,curr_yr.sales_cnt AS ⛓⚓curr_yr_cnt ✖⚓,curr_yr.sales_cnt ⚫-⚫ prev_yr.sales_cnt AS ⛓⚓sales_cnt_diff ✖⚓,curr_yr.sales_amt ⚫-⚫ prev_yr.sales_amt AS ⛓⚓sales_amt_diff❌FROM⚓ all_sales curr_yr✖⚓,all_sales prev_yr❌WHERE⚓ curr_yr.brand_id ⚫=⚫ prev_yr.brand_id ✖AND⚓ curr_yr.class_id ⚫=⚫ prev_yr.class_id ✖AND⚓ curr_yr.category_id ⚫=⚫ prev_yr.category_id ✖AND⚓ curr_yr.manufact_id ⚫=⚫ prev_yr.manufact_id ✖AND⚓ curr_yr.year ⚫=⚫ 2000 ✖AND⚓ prev_yr.year ⚫=⚫ 2000 ⚫-⚫ 1 ✖AND⚓ CAST(➕curr_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫/⚫ CAST(➕prev_yr.sales_cnt AS DECIMAL(➕17✖⚓,2➖)➖) ⚫<⚫ 0.9❌ORDER⚓ BY sales_cnt_diff❌limit⚓ 100"""

        expected = """WITH all_sales AS(
    SELECT d.year 
          ,i.brand_id 
          ,i.class_id 
          ,i.category_id 
          ,i.manufact_id 
          ,SUM(sales_cnt) AS 
          sales_cnt 
          ,SUM(sales_amt) AS 
          sales_amt
      FROM(SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,cs.quantity - COALESCE(cr.return_quantity
                  ,0) AS 
                 sales_cnt 
                 ,cs.ext_sales_price - COALESCE(cr.return_amount
                  ,0.0) AS 
                 sales_amt 
             FROM catalog_sales cs 
             JOIN item i 
               ON i.item_sk = cs.item_sk
             JOIN date_dim d 
               ON d.date_sk = cs.sold_date_sk
             LEFT 
             JOIN catalog_returns cr 
               ON(cs.order_number = cr.order_number
               AND cs.item_sk = cr.item_sk)
            WHERE i.category = 'Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ss.quantity - COALESCE(sr.return_quantity
                  ,0) AS 
                 sales_cnt 
                 ,ss.ext_sales_price - COALESCE(sr.return_amt
                  ,0.0) AS 
                 sales_amt 
             FROM store_sales ss 
             JOIN item i 
               ON i.item_sk = ss.item_sk
             JOIN date_dim d 
               ON d.date_sk = ss.sold_date_sk
             LEFT 
             JOIN store_returns sr 
               ON(ss.ticket_number = sr.ticket_number
               AND ss.item_sk = sr.item_sk)
            WHERE i.category = 'Shoes'
            UNION
           SELECT d.year 
                 ,i.brand_id 
                 ,i.class_id 
                 ,i.category_id 
                 ,i.manufact_id 
                 ,ws.quantity - COALESCE(wr.return_quantity
                  ,0) AS 
                 sales_cnt 
                 ,ws.ext_sales_price - COALESCE(wr.return_amt
                  ,0.0) AS 
                 sales_amt 
             FROM web_sales ws 
             JOIN item i 
               ON i.item_sk = ws.item_sk
             JOIN date_dim d 
               ON d.date_sk = ws.sold_date_sk
             LEFT 
             JOIN web_returns wr 
               ON(ws.order_number = wr.order_number
               AND ws.item_sk = wr.item_sk)
            WHERE i.category = 'Shoes') sales_detail
     GROUP BY d.year
          ,i.brand_id
          ,i.class_id
          ,i.category_id
          ,i.manufact_id
          )
SELECT prev_yr.d_year AS 
      prev_year 
      ,curr_yr.d_year AS 
      year 
      ,curr_yr.i_brand_id 
      ,curr_yr.i_class_id 
      ,curr_yr.i_category_id 
      ,curr_yr.i_manufact_id 
      ,prev_yr.sales_cnt AS 
      prev_yr_cnt 
      ,curr_yr.sales_cnt AS 
      curr_yr_cnt 
      ,curr_yr.sales_cnt - prev_yr.sales_cnt AS
      sales_cnt_diff 
      ,curr_yr.sales_amt - prev_yr.sales_amt AS
      sales_amt_diff
  FROM all_sales curr_yr
      ,all_sales prev_yr
 WHERE curr_yr.brand_id = prev_yr.brand_id
   AND curr_yr.class_id = prev_yr.class_id
   AND curr_yr.category_id = prev_yr.category_id
   AND curr_yr.manufact_id = prev_yr.manufact_id
   AND curr_yr.year = 2000
   AND prev_yr.year = 2000 - 1
   AND CAST(curr_yr.sales_cnt AS DECIMAL(17
                 ,2)) / CAST(prev_yr.sales_cnt AS DECIMAL(17
                 ,2)) < 0.9
 ORDER BY sales_cnt_diff
 limit 100"""

        self.assertEqual(expected, _process_data(marked_sql, "comfortable"))

    def test_sample_simple_comfortable(self):
        marked_sql = """SELECT⚓ p1✖⚓,p2✖⚓,count(➕*➖) as ⛓⚓numorders❌FROM⚓(➕SELECT⚓ op1.OrderID✖⚓,op1.ProductID as ⛓⚓p1✖⚓,op2.ProductID as ⛓⚓p2 ❌FROM⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op1 ❌JOIN⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op2 ❌ON⚓ op1.OrderID ⚫=⚫ op2.OrderID ✖AND⚓ op1.ProductID ⚫>⚫ op2.ProductID➖)combinations❌GROUP⚓ BY p1✖⚓,p2"""

        expected = """SELECT p1
      ,p2
      ,count(*) as 
      numorders
  FROM(SELECT op1.OrderID
             ,op1.ProductID as 
             p1
             ,op2.ProductID as 
             p2 
         FROM(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op1 
         JOIN(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op2 
           ON op1.OrderID = op2.OrderID
          AND op1.ProductID > op2.ProductID)combinations
 GROUP BY p1
      ,p2"""

        self.assertEqual(expected, _process_data(marked_sql, "comfortable"))

    def test_sample_simple_compact(self):
        marked_sql = """SELECT⚓ p1✖⚓,p2✖⚓,count(➕*➖) as ⛓⚓numorders❌FROM⚓(➕SELECT⚓ op1.OrderID✖⚓,op1.ProductID as ⛓⚓p1✖⚓,op2.ProductID as ⛓⚓p2 ❌FROM⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op1 ❌JOIN⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op2 ❌ON⚓ op1.OrderID ⚫=⚫ op2.OrderID ✖AND⚓ op1.ProductID ⚫>⚫ op2.ProductID➖)combinations❌GROUP⚓ BY p1✖⚓,p2"""

        expected = """SELECT p1
      ,p2
      ,count(*) as numorders
  FROM(SELECT op1.OrderID
             ,op1.ProductID as p1
             ,op2.ProductID as p2 
         FROM(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op1 
         JOIN(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op2 
           ON op1.OrderID=op2.OrderID
          AND op1.ProductID>op2.ProductID)combinations
 GROUP BY p1
      ,p2"""

        self.assertEqual(expected, _process_data(marked_sql, "compact"))

    def test_sample_simple_balanced(self):
        marked_sql = """SELECT⚓ p1✖⚓,p2✖⚓,count(➕*➖) as ⛓⚓numorders❌FROM⚓(➕SELECT⚓ op1.OrderID✖⚓,op1.ProductID as ⛓⚓p1✖⚓,op2.ProductID as ⛓⚓p2 ❌FROM⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op1 ❌JOIN⚓(➕SELECT⚓ DISTINCT OrderID✖⚓,ProductID ❌FROM⚓ OrderLines➖)op2 ❌ON⚓ op1.OrderID ⚫=⚫ op2.OrderID ✖AND⚓ op1.ProductID ⚫>⚫ op2.ProductID➖)combinations❌GROUP⚓ BY p1✖⚓,p2"""

        expected = """SELECT p1
      ,p2
      ,count(*) as numorders
  FROM(SELECT op1.OrderID
             ,op1.ProductID as p1
             ,op2.ProductID as p2 
         FROM(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op1 
         JOIN(SELECT DISTINCT OrderID
                    ,ProductID 
                FROM OrderLines)op2 
           ON op1.OrderID = op2.OrderID
          AND op1.ProductID > op2.ProductID)combinations
 GROUP BY p1
      ,p2"""

        self.assertEqual(expected, _process_data(marked_sql, "balanced"))

    def test_sample_complex_balanced(self):
        marked_sql = """WITH← date_list AS(➕❌ SELECT⚓ CAST(➕'2021-05-24' AS date➖) AS ⛓⚓OrderDate ❌UNION⚓❌ SELECT⚓ CAST(➕'2021-05-25' AS date➖) AS ⛓⚓OrderDate⛓⚓➖)❌,unique_orderedouters AS(➕❌ select⚓ PURCHASEORDERLINEID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27229 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_purchase as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27033 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_item as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 114364 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_parent as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 34559 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,purchase_con AS(➕❌ select⚓ DISTINCT CONCAT(➕pol.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,dl.OrderDate as ⛓⚓search_date✖⚓,pol.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sinv.ORDERID✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of the Voucher' as ⛓⚓ParentLevel_Desc ❌from⚓ PURCHASING_PURCHASEORDERLINES pol ❌inner⚓ ❌join⚓ SALES_ORDERS sord ❌on⚓ sord.ORDERID ⚫=⚫ pol.PURCHASEORDERID ✖and⚓ pol.PURCHASEORDERID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null ❌inner⚓ ❌join⚓ SALES_INVOICES sinv ❌on⚓ sinv.ORDERID ⚫=⚫ sord.ORDERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERS AS scus ❌ON⚓ scus.CUSTOMERID ⚫=⚫ sinv.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERTRANSACTIONS AS sct ❌ON⚓ sct.CUSTOMERID ⚫=⚫ scus.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_INVOICELINES AS sinvl ❌ON⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌INNER⚓ ❌join⚓ unique_purchase up ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌LEFT⚓ ❌join⚓ unique_item ui ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌INNER⚓ ❌JOIN⚓ date_list dl ❌ON⚓ pol.LASTRECEIPTDATE⚫<=⚫ dl.OrderDate ✖AND⚓(➕ pol.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ pol.LASTEDITEDWHEN is NULL➖) ✖AND⚓ sinv.INVOICEDATE ⚫<=⚫ dl.OrderDate ✖AND⚓(➕ sinv.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ sinv.LASTEDITEDWHEN is NULL➖) ❌where⚓ pol.STOCKITEMID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null⛓⚓➖)❌,rule AS(➕❌ select⚓ distinct CONCAT(➕orderedouters.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,orderedouters.OrderDate as ⛓⚓search_date✖⚓,orderedouters.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of Salesperson' as ⛓⚓ParentLevel_Desc ❌from⚓ unique_orderedouters orderedouters ❌inner⚓ ❌join⚓ SALES_CUSTOMERTRANSACTIONS sct ❌on⚓ orderedouters.PURCHASEORDERLINEID ⚫=⚫ sct.PURCHASEORDERLINEID ❌inner⚓ ❌join⚓ SALES_INVOICELINES sinvl ❌on⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌inner⚓ ❌join⚓ unique_purchase up ❌on⚓ up.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ❌LEFT⚓ ❌join⚓ unique_item up ❌on⚓ ui.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ✖and⚓ ui.OrderDate ⚫=⚫ orderedouters.OrderDate ❌where⚓ sct.ISFINALIZED in(➕'TRUE'➖)⛓⚓➖)❌,direct_map AS(➕❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ purchase_con ❌UNION⚓❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ rule ❌where⚓ ID not in(➕ select⚓ distinct ID ❌from⚓ purchase_con➖)⛓⚓➖)❌,parent_child AS(➕❌ select⚓ Distinct upar.OrderDate AS ⛓⚓search_date✖⚓,upar.STOCKITEMIDas ⛓⚓PolID✖⚓,UPPER(➕cls.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,cln.ORDEREDOUTERS as ⛓⚓outersid✖⚓,upar.ORDEREDOUTERS as ⛓⚓parent_PolID✖⚓,UPPER(➕pls.ORDEREDOUTERS➖) as ⛓⚓parent_outersname✖⚓,pln.ORDEREDOUTERS as ⛓⚓parent_outersid ❌from⚓ unique_parent upar ❌left⚓ ❌join⚓ unique_purchase uprc ❌on⚓ uprc.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uprc.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_purchase uprcs ❌on⚓ pls.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uprcs.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uit ❌on⚓ cln.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uit.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uitem ❌on⚓ uitem.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uitem.OrderDate ⚫=⚫ upar.OrderDate⛓⚓➖)❌,navigate_hierarchy as(➕❌ SELECT⚓ search_date✖⚓,PolID✖⚓,outersid✖⚓,cast(➕outersname as varchar➖) as ⛓⚓outersname✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,cast(➕parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,1 as ⛓⚓ParentLevel✖⚓,concat(➕ cast(➕outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕parent_outersname as varchar➖)➖) as ⛓⚓outersnameline ❌from⚓ parent_child ❌UNION ALL⚓❌ SELECT⚓ c.search_date✖⚓,c.PolID✖⚓,c.outersid✖⚓,cast(➕c.outersname as varchar➖) as ⛓⚓outersname✖⚓,p.parent_PolID✖⚓,p.parent_outersid✖⚓,cast(➕p.parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,p.ParentLevel ⚫+⚫ 1✖⚓,cast(➕ CONCAT(➕ cast(➕c.outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕p.outersnameline as varchar➖)➖) AS varchar➖) as ⛓⚓outersnameline ❌from⚓ parent_child C ❌join⚓ navigate_hierarchy p ❌on⚓ p.PolID ⚫=⚫ c.parent_PolID ✖AND⚓ c.PolID ⚫<>⚫ c.parent_PolID ✖and⚓ p.ParentLevel ⚫<=⚫ 40 ✖AND⚓ p.search_date ⚫=⚫ c.search_date ❌where⚓ p.ParentLevel ⚫<=⚫ 40⛓⚓➖)❌,TEST_OUTPUT AS(➕❌ select⚓ DISTINCT ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc✖⚓,PolID as ⛓⚓parent_PolID✖⚓,outersid as ⛓⚓parent_outersid✖⚓,outersname as ⛓⚓parent_outersname✖⚓,outersname as ⛓⚓outersnameline ❌from⚓ direct_mapping ❌UNION⚓❌ select⚓ CONCAT(➕dm.orderedouters✖⚓,'-'✖⚓,nh.Parentlevel➖) as ⛓⚓ID✖⚓,dm.search_date✖⚓,dm.orderedouters✖⚓,dm.PAYMENTMETHODID✖⚓,dm.TRANSACTIONDATE✖⚓,nh.PolID✖⚓,nh.outersid✖⚓,nh.outersname✖⚓,dm.STOCKITEMID✖⚓,nh.ParentLevel✖⚓,❌CASE⚓➕ ❌WHEN⚓ nh.ParentLevel ⚫=⚫ 1 ❌THEN⚓ 'Immediate Parent of Salesperson Issuer' ❌ELSE⚓ concat(➕ 'Higher Level Parent ('✖⚓,nh.ParentLevel✖⚓,') Levels above the Salesperson Issuer'➖) ➖❌END⚓ as ⛓⚓ParentLevel_Desc✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,parent_outersname✖⚓,outersnameline as ⛓⚓outersnameline ❌from⚓ navigate_hierarchy nh ❌inner⚓ ❌join⚓ direct_map dm ❌on⚓ dm.PolID ⚫=⚫ nh.PolID⛓⚓➖)❌SELECT⚓ *❌FROM⚓ TEST_OUTPUT ❌group⚓ by search_date"""

        expected = """WITH date_list AS(
    SELECT CAST('2021-05-24' AS date) AS OrderDate 
     UNION
    SELECT CAST('2021-05-25' AS date) AS OrderDate)
,unique_orderedouters AS(
    select PURCHASEORDERLINEID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 27229
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1)
,unique_purchase as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 27033
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1)
,unique_item as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 114364
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1)
,unique_parent as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 34559
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1)
,purchase_con AS(
    select DISTINCT CONCAT(pol.ORDEREDOUTERS
              ,'-0') as ID
          ,dl.OrderDate as search_date
          ,pol.ORDEREDOUTERS as orderedouters
          ,sinv.ORDERID
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as PolID
          ,ui.ORDEREDOUTERS as outersid
          ,UPPER(up.ORDEREDOUTERS) as outersname
          ,sinvl.STOCKITEMID
          ,0 as ParentLevel
          ,'Direct Issuer of the Voucher' as ParentLevel_Desc 
      from PURCHASING_PURCHASEORDERLINES pol 
     inner 
      join SALES_ORDERS sord 
        on sord.ORDERID = pol.PURCHASEORDERID
       and pol.PURCHASEORDERID = 35
       and pol.STOCKITEMID = 1
       and sord.SALESPERSONPERSONID = 1
       and sord.CONTACTPERSONID = 1
       and sord.DELIVERYINSTRUCTIONS = 'SHR'
       AND pol.ORDEREDOUTERS IS NOT NULL 
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null 
     inner 
      join SALES_INVOICES sinv 
        on sinv.ORDERID = sord.ORDERID
     INNER 
      JOIN SALES_CUSTOMERS AS scus 
        ON scus.CUSTOMERID = sinv.CUSTOMERID
     INNER 
      JOIN SALES_CUSTOMERTRANSACTIONS AS sct 
        ON sct.CUSTOMERID = scus.CUSTOMERID
     INNER 
      JOIN SALES_INVOICELINES AS sinvl 
        ON sinvl.INVOICEID = sct.INVOICEID
     INNER 
      join unique_purchase up 
        on ui.STOCKITEMID = sinvl.STOCKITEMID
      LEFT 
      join unique_item ui 
        on ui.STOCKITEMID = sinvl.STOCKITEMID
     INNER 
      JOIN date_list dl 
        ON pol.LASTRECEIPTDATE <= dl.OrderDate
       AND( pol.LASTEDITEDWHEN > dl.OrderDate
            OR pol.LASTEDITEDWHEN is NULL) 
       AND sinv.INVOICEDATE <= dl.OrderDate
       AND( sinv.LASTEDITEDWHEN > dl.OrderDate
            OR sinv.LASTEDITEDWHEN is NULL) 
     where pol.STOCKITEMID = 35
       and pol.STOCKITEMID = 1
       AND pol.ORDEREDOUTERS IS NOT NULL 
       and sord.SALESPERSONPERSONID = 1
       and sord.CONTACTPERSONID = 1
       and sord.DELIVERYINSTRUCTIONS = 'SHR'
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null)
,rule AS(
    select distinct CONCAT(orderedouters.ORDEREDOUTERS
              ,'-0') as ID
          ,orderedouters.OrderDate as search_date
          ,orderedouters.ORDEREDOUTERS as orderedouters
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as PolID
          ,ui.ORDEREDOUTERS as outersid
          ,UPPER(up.ORDEREDOUTERS) as outersname
          ,sinvl.STOCKITEMID
          ,0 as ParentLevel
          ,'Direct Issuer of Salesperson' as ParentLevel_Desc 
      from unique_orderedouters orderedouters 
     inner 
      join SALES_CUSTOMERTRANSACTIONS sct 
        on orderedouters.PURCHASEORDERLINEID = sct.PURCHASEORDERLINEID
     inner 
      join SALES_INVOICELINES sinvl 
        on sinvl.INVOICEID = sct.INVOICEID
     inner 
      join unique_purchase up 
        on up.PURCHASEORDERLINEID = sinvl.PURCHASEORDERLINEID
      LEFT 
      join unique_item up 
        on ui.PURCHASEORDERLINEID = sinvl.PURCHASEORDERLINEID
       and ui.OrderDate = orderedouters.OrderDate
     where sct.ISFINALIZED in('TRUE'))
,direct_map AS(
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from purchase_con 
     UNION
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from rule 
     where ID not in( select distinct ID 
                         from purchase_con))
,parent_child AS(
    select Distinct upar.OrderDate AS search_date
          ,upar.STOCKITEMIDas PolID
          ,UPPER(cls.ORDEREDOUTERS) as outersname
          ,cln.ORDEREDOUTERS as outersid
          ,upar.ORDEREDOUTERS as parent_PolID
          ,UPPER(pls.ORDEREDOUTERS) as parent_outersname
          ,pln.ORDEREDOUTERS as parent_outersid 
      from unique_parent upar 
      left 
      join unique_purchase uprc 
        on uprc.STOCKITEMID = upar.STOCKITEMID
       and uprc.OrderDate = upar.OrderDate
      left 
      join unique_purchase uprcs 
        on pls.STOCKITEMID = upar.ORDEREDOUTERS
       and uprcs.OrderDate = upar.OrderDate
      left 
      join unique_item uit 
        on cln.STOCKITEMID = upar.STOCKITEMID
       and uit.OrderDate = upar.OrderDate
      left 
      join unique_item uitem 
        on uitem.STOCKITEMID = upar.ORDEREDOUTERS
       and uitem.OrderDate = upar.OrderDate)
,navigate_hierarchy as(
    SELECT search_date
          ,PolID
          ,outersid
          ,cast(outersname as varchar) as outersname
          ,parent_PolID
          ,parent_outersid
          ,cast(parent_outersname as varchar) as parent_outersname
          ,1 as ParentLevel
          ,concat( cast(outersname as varchar)
                             ,' -> '
                             ,cast(parent_outersname as varchar)) as outersnameline 
      from parent_child 
   UNION ALL
    SELECT c.search_date
          ,c.PolID
          ,c.outersid
          ,cast(c.outersname as varchar) as outersname
          ,p.parent_PolID
          ,p.parent_outersid
          ,cast(p.parent_outersname as varchar) as parent_outersname
          ,p.ParentLevel + 1
          ,cast( CONCAT( cast(c.outersname as varchar)
                  ,' -> '
                  ,cast(p.outersnameline as varchar)) AS varchar) as outersnameline 
      from parent_child C 
      join navigate_hierarchy p 
        on p.PolID = c.parent_PolID
       AND c.PolID <> c.parent_PolID
       and p.ParentLevel <= 40
       AND p.search_date = c.search_date
     where p.ParentLevel <= 40)
,TEST_OUTPUT AS(
    select DISTINCT ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc
          ,PolID as parent_PolID
          ,outersid as parent_outersid
          ,outersname as parent_outersname
          ,outersname as outersnameline 
      from direct_mapping 
     UNION
    select CONCAT(dm.orderedouters
                             ,'-'
                             ,nh.Parentlevel) as ID
          ,dm.search_date
          ,dm.orderedouters
          ,dm.PAYMENTMETHODID
          ,dm.TRANSACTIONDATE
          ,nh.PolID
          ,nh.outersid
          ,nh.outersname
          ,dm.STOCKITEMID
          ,nh.ParentLevel
          ,
      CASE 
                         WHEN nh.ParentLevel = 1
                         THEN 'Immediate Parent of Salesperson Issuer' 
                         ELSE concat( 'Higher Level Parent ('
                  ,nh.ParentLevel
                  ,') Levels above the Salesperson Issuer') 
       END as ParentLevel_Desc
          ,parent_PolID
          ,parent_outersid
          ,parent_outersname
          ,outersnameline as outersnameline 
      from navigate_hierarchy nh 
     inner 
      join direct_map dm 
        on dm.PolID = nh.PolID)
SELECT *
  FROM TEST_OUTPUT 
 group by search_date"""

        self.assertEqual(expected, _process_data(marked_sql, "balanced"))

    def test_sample_complex_comfortable(self):
        marked_sql = """WITH← date_list AS(➕❌ SELECT⚓ CAST(➕'2021-05-24' AS date➖) AS ⛓⚓OrderDate ❌UNION⚓❌ SELECT⚓ CAST(➕'2021-05-25' AS date➖) AS ⛓⚓OrderDate⛓⚓➖)❌,unique_orderedouters AS(➕❌ select⚓ PURCHASEORDERLINEID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27229 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_purchase as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27033 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_item as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 114364 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_parent as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 34559 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,purchase_con AS(➕❌ select⚓ DISTINCT CONCAT(➕pol.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,dl.OrderDate as ⛓⚓search_date✖⚓,pol.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sinv.ORDERID✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of the Voucher' as ⛓⚓ParentLevel_Desc ❌from⚓ PURCHASING_PURCHASEORDERLINES pol ❌inner⚓ ❌join⚓ SALES_ORDERS sord ❌on⚓ sord.ORDERID ⚫=⚫ pol.PURCHASEORDERID ✖and⚓ pol.PURCHASEORDERID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null ❌inner⚓ ❌join⚓ SALES_INVOICES sinv ❌on⚓ sinv.ORDERID ⚫=⚫ sord.ORDERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERS AS scus ❌ON⚓ scus.CUSTOMERID ⚫=⚫ sinv.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERTRANSACTIONS AS sct ❌ON⚓ sct.CUSTOMERID ⚫=⚫ scus.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_INVOICELINES AS sinvl ❌ON⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌INNER⚓ ❌join⚓ unique_purchase up ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌LEFT⚓ ❌join⚓ unique_item ui ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌INNER⚓ ❌JOIN⚓ date_list dl ❌ON⚓ pol.LASTRECEIPTDATE⚫<=⚫ dl.OrderDate ✖AND⚓(➕ pol.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ pol.LASTEDITEDWHEN is NULL➖) ✖AND⚓ sinv.INVOICEDATE ⚫<=⚫ dl.OrderDate ✖AND⚓(➕ sinv.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ sinv.LASTEDITEDWHEN is NULL➖) ❌where⚓ pol.STOCKITEMID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null⛓⚓➖)❌,rule AS(➕❌ select⚓ distinct CONCAT(➕orderedouters.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,orderedouters.OrderDate as ⛓⚓search_date✖⚓,orderedouters.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of Salesperson' as ⛓⚓ParentLevel_Desc ❌from⚓ unique_orderedouters orderedouters ❌inner⚓ ❌join⚓ SALES_CUSTOMERTRANSACTIONS sct ❌on⚓ orderedouters.PURCHASEORDERLINEID ⚫=⚫ sct.PURCHASEORDERLINEID ❌inner⚓ ❌join⚓ SALES_INVOICELINES sinvl ❌on⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌inner⚓ ❌join⚓ unique_purchase up ❌on⚓ up.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ❌LEFT⚓ ❌join⚓ unique_item up ❌on⚓ ui.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ✖and⚓ ui.OrderDate ⚫=⚫ orderedouters.OrderDate ❌where⚓ sct.ISFINALIZED in(➕'TRUE'➖)⛓⚓➖)❌,direct_map AS(➕❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ purchase_con ❌UNION⚓❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ rule ❌where⚓ ID not in(➕ select⚓ distinct ID ❌from⚓ purchase_con➖)⛓⚓➖)❌,parent_child AS(➕❌ select⚓ Distinct upar.OrderDate AS ⛓⚓search_date✖⚓,upar.STOCKITEMIDas ⛓⚓PolID✖⚓,UPPER(➕cls.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,cln.ORDEREDOUTERS as ⛓⚓outersid✖⚓,upar.ORDEREDOUTERS as ⛓⚓parent_PolID✖⚓,UPPER(➕pls.ORDEREDOUTERS➖) as ⛓⚓parent_outersname✖⚓,pln.ORDEREDOUTERS as ⛓⚓parent_outersid ❌from⚓ unique_parent upar ❌left⚓ ❌join⚓ unique_purchase uprc ❌on⚓ uprc.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uprc.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_purchase uprcs ❌on⚓ pls.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uprcs.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uit ❌on⚓ cln.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uit.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uitem ❌on⚓ uitem.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uitem.OrderDate ⚫=⚫ upar.OrderDate⛓⚓➖)❌,navigate_hierarchy as(➕❌ SELECT⚓ search_date✖⚓,PolID✖⚓,outersid✖⚓,cast(➕outersname as varchar➖) as ⛓⚓outersname✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,cast(➕parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,1 as ⛓⚓ParentLevel✖⚓,concat(➕ cast(➕outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕parent_outersname as varchar➖)➖) as ⛓⚓outersnameline ❌from⚓ parent_child ❌UNION ALL⚓❌ SELECT⚓ c.search_date✖⚓,c.PolID✖⚓,c.outersid✖⚓,cast(➕c.outersname as varchar➖) as ⛓⚓outersname✖⚓,p.parent_PolID✖⚓,p.parent_outersid✖⚓,cast(➕p.parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,p.ParentLevel ⚫+⚫ 1✖⚓,cast(➕ CONCAT(➕ cast(➕c.outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕p.outersnameline as varchar➖)➖) AS varchar➖) as ⛓⚓outersnameline ❌from⚓ parent_child C ❌join⚓ navigate_hierarchy p ❌on⚓ p.PolID ⚫=⚫ c.parent_PolID ✖AND⚓ c.PolID ⚫<>⚫ c.parent_PolID ✖and⚓ p.ParentLevel ⚫<=⚫ 40 ✖AND⚓ p.search_date ⚫=⚫ c.search_date ❌where⚓ p.ParentLevel ⚫<=⚫ 40⛓⚓➖)❌,TEST_OUTPUT AS(➕❌ select⚓ DISTINCT ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc✖⚓,PolID as ⛓⚓parent_PolID✖⚓,outersid as ⛓⚓parent_outersid✖⚓,outersname as ⛓⚓parent_outersname✖⚓,outersname as ⛓⚓outersnameline ❌from⚓ direct_mapping ❌UNION⚓❌ select⚓ CONCAT(➕dm.orderedouters✖⚓,'-'✖⚓,nh.Parentlevel➖) as ⛓⚓ID✖⚓,dm.search_date✖⚓,dm.orderedouters✖⚓,dm.PAYMENTMETHODID✖⚓,dm.TRANSACTIONDATE✖⚓,nh.PolID✖⚓,nh.outersid✖⚓,nh.outersname✖⚓,dm.STOCKITEMID✖⚓,nh.ParentLevel✖⚓,❌CASE⚓➕ ❌WHEN⚓ nh.ParentLevel ⚫=⚫ 1 ❌THEN⚓ 'Immediate Parent of Salesperson Issuer' ❌ELSE⚓ concat(➕ 'Higher Level Parent ('✖⚓,nh.ParentLevel✖⚓,') Levels above the Salesperson Issuer'➖) ➖❌END⚓ as ⛓⚓ParentLevel_Desc✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,parent_outersname✖⚓,outersnameline as ⛓⚓outersnameline ❌from⚓ navigate_hierarchy nh ❌inner⚓ ❌join⚓ direct_map dm ❌on⚓ dm.PolID ⚫=⚫ nh.PolID⛓⚓➖)❌SELECT⚓ *❌FROM⚓ TEST_OUTPUT ❌group⚓ by search_date"""

        expected = """WITH date_list AS(
    SELECT CAST('2021-05-24' AS date) AS 
          OrderDate 
     UNION
    SELECT CAST('2021-05-25' AS date) AS 
          OrderDate
          )
,unique_orderedouters AS(
    select PURCHASEORDERLINEID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 27229
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1
          )
,unique_purchase as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 27033
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1
          )
,unique_item as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 114364
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1
          )
,unique_parent as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE <= OrderDate
       AND( a.LASTEDITEDWHEN > OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID = 34559
       and a.STOCKITEMID = 1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC) = 1
          )
,purchase_con AS(
    select DISTINCT CONCAT(pol.ORDEREDOUTERS
              ,'-0') as 
          ID
          ,dl.OrderDate as 
          search_date
          ,pol.ORDEREDOUTERS as 
          orderedouters
          ,sinv.ORDERID
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as 
          PolID
          ,ui.ORDEREDOUTERS as 
          outersid
          ,UPPER(up.ORDEREDOUTERS) as 
          outersname
          ,sinvl.STOCKITEMID
          ,0 as 
          ParentLevel
          ,'Direct Issuer of the Voucher' as 
          ParentLevel_Desc 
      from PURCHASING_PURCHASEORDERLINES pol 
     inner 
      join SALES_ORDERS sord 
        on sord.ORDERID = pol.PURCHASEORDERID
       and pol.PURCHASEORDERID = 35
       and pol.STOCKITEMID = 1
       and sord.SALESPERSONPERSONID = 1
       and sord.CONTACTPERSONID = 1
       and sord.DELIVERYINSTRUCTIONS = 'SHR'
       AND pol.ORDEREDOUTERS IS NOT NULL 
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null 
     inner 
      join SALES_INVOICES sinv 
        on sinv.ORDERID = sord.ORDERID
     INNER 
      JOIN SALES_CUSTOMERS AS scus 
        ON scus.CUSTOMERID = sinv.CUSTOMERID
     INNER 
      JOIN SALES_CUSTOMERTRANSACTIONS AS sct 
        ON sct.CUSTOMERID = scus.CUSTOMERID
     INNER 
      JOIN SALES_INVOICELINES AS sinvl 
        ON sinvl.INVOICEID = sct.INVOICEID
     INNER 
      join unique_purchase up 
        on ui.STOCKITEMID = sinvl.STOCKITEMID
      LEFT 
      join unique_item ui 
        on ui.STOCKITEMID = sinvl.STOCKITEMID
     INNER 
      JOIN date_list dl 
        ON pol.LASTRECEIPTDATE <= dl.OrderDate
       AND( pol.LASTEDITEDWHEN > dl.OrderDate
            OR pol.LASTEDITEDWHEN is NULL) 
       AND sinv.INVOICEDATE <= dl.OrderDate
       AND( sinv.LASTEDITEDWHEN > dl.OrderDate
            OR sinv.LASTEDITEDWHEN is NULL) 
     where pol.STOCKITEMID = 35
       and pol.STOCKITEMID = 1
       AND pol.ORDEREDOUTERS IS NOT NULL 
       and sord.SALESPERSONPERSONID = 1
       and sord.CONTACTPERSONID = 1
       and sord.DELIVERYINSTRUCTIONS = 'SHR'
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null
          )
,rule AS(
    select distinct CONCAT(orderedouters.ORDEREDOUTERS
              ,'-0') as 
          ID
          ,orderedouters.OrderDate as 
          search_date
          ,orderedouters.ORDEREDOUTERS as 
          orderedouters
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as 
          PolID
          ,ui.ORDEREDOUTERS as 
          outersid
          ,UPPER(up.ORDEREDOUTERS) as 
          outersname
          ,sinvl.STOCKITEMID
          ,0 as 
          ParentLevel
          ,'Direct Issuer of Salesperson' as 
          ParentLevel_Desc 
      from unique_orderedouters orderedouters 
     inner 
      join SALES_CUSTOMERTRANSACTIONS sct 
        on orderedouters.PURCHASEORDERLINEID = sct.PURCHASEORDERLINEID
     inner 
      join SALES_INVOICELINES sinvl 
        on sinvl.INVOICEID = sct.INVOICEID
     inner 
      join unique_purchase up 
        on up.PURCHASEORDERLINEID = sinvl.PURCHASEORDERLINEID
      LEFT 
      join unique_item up 
        on ui.PURCHASEORDERLINEID = sinvl.PURCHASEORDERLINEID
       and ui.OrderDate = orderedouters.OrderDate
     where sct.ISFINALIZED in('TRUE')
          )
,direct_map AS(
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from purchase_con 
     UNION
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from rule 
     where ID not in( select distinct ID 
                         from purchase_con)
          )
,parent_child AS(
    select Distinct upar.OrderDate AS 
          search_date
          ,upar.STOCKITEMIDas 
          PolID
          ,UPPER(cls.ORDEREDOUTERS) as 
          outersname
          ,cln.ORDEREDOUTERS as 
          outersid
          ,upar.ORDEREDOUTERS as 
          parent_PolID
          ,UPPER(pls.ORDEREDOUTERS) as 
          parent_outersname
          ,pln.ORDEREDOUTERS as 
          parent_outersid 
      from unique_parent upar 
      left 
      join unique_purchase uprc 
        on uprc.STOCKITEMID = upar.STOCKITEMID
       and uprc.OrderDate = upar.OrderDate
      left 
      join unique_purchase uprcs 
        on pls.STOCKITEMID = upar.ORDEREDOUTERS
       and uprcs.OrderDate = upar.OrderDate
      left 
      join unique_item uit 
        on cln.STOCKITEMID = upar.STOCKITEMID
       and uit.OrderDate = upar.OrderDate
      left 
      join unique_item uitem 
        on uitem.STOCKITEMID = upar.ORDEREDOUTERS
       and uitem.OrderDate = upar.OrderDate
          )
,navigate_hierarchy as(
    SELECT search_date
          ,PolID
          ,outersid
          ,cast(outersname as varchar) as 
          outersname
          ,parent_PolID
          ,parent_outersid
          ,cast(parent_outersname as varchar) as 
          parent_outersname
          ,1 as 
          ParentLevel
          ,concat( cast(outersname as varchar)
                             ,' -> '
                             ,cast(parent_outersname as varchar)) as 
          outersnameline 
      from parent_child 
   UNION ALL
    SELECT c.search_date
          ,c.PolID
          ,c.outersid
          ,cast(c.outersname as varchar) as 
          outersname
          ,p.parent_PolID
          ,p.parent_outersid
          ,cast(p.parent_outersname as varchar) as 
          parent_outersname
          ,p.ParentLevel + 1
          ,cast( CONCAT( cast(c.outersname as varchar)
                  ,' -> '
                  ,cast(p.outersnameline as varchar)) AS varchar) as 
          outersnameline 
      from parent_child C 
      join navigate_hierarchy p 
        on p.PolID = c.parent_PolID
       AND c.PolID <> c.parent_PolID
       and p.ParentLevel <= 40
       AND p.search_date = c.search_date
     where p.ParentLevel <= 40
          )
,TEST_OUTPUT AS(
    select DISTINCT ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc
          ,PolID as 
          parent_PolID
          ,outersid as 
          parent_outersid
          ,outersname as 
          parent_outersname
          ,outersname as 
          outersnameline 
      from direct_mapping 
     UNION
    select CONCAT(dm.orderedouters
                             ,'-'
                             ,nh.Parentlevel) as 
          ID
          ,dm.search_date
          ,dm.orderedouters
          ,dm.PAYMENTMETHODID
          ,dm.TRANSACTIONDATE
          ,nh.PolID
          ,nh.outersid
          ,nh.outersname
          ,dm.STOCKITEMID
          ,nh.ParentLevel
          ,
      CASE 
                         WHEN nh.ParentLevel = 1
                         THEN 'Immediate Parent of Salesperson Issuer' 
                         ELSE concat( 'Higher Level Parent ('
                  ,nh.ParentLevel
                  ,') Levels above the Salesperson Issuer') 
       END as 
          ParentLevel_Desc
          ,parent_PolID
          ,parent_outersid
          ,parent_outersname
          ,outersnameline as 
          outersnameline 
      from navigate_hierarchy nh 
     inner 
      join direct_map dm 
        on dm.PolID = nh.PolID
          )
SELECT *
  FROM TEST_OUTPUT 
 group by search_date"""

        self.assertEqual(expected, _process_data(marked_sql, "comfortable"))

    def test_sample_complex_compact(self):
        marked_sql = """WITH← date_list AS(➕❌ SELECT⚓ CAST(➕'2021-05-24' AS date➖) AS ⛓⚓OrderDate ❌UNION⚓❌ SELECT⚓ CAST(➕'2021-05-25' AS date➖) AS ⛓⚓OrderDate⛓⚓➖)❌,unique_orderedouters AS(➕❌ select⚓ PURCHASEORDERLINEID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27229 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_purchase as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 27033 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_item as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 114364 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,unique_parent as(➕❌ select⚓ STOCKITEMID✖⚓,ORDEREDOUTERS✖⚓,dl.OrderDate ❌from⚓ PURCHASING_PURCHASEORDERLINES a ❌JOIN⚓ date_list dl ❌ON⚓ a.LASTRECEIPTDATE ⚫<=⚫ OrderDate ✖AND⚓(➕ a.LASTEDITEDWHEN ⚫>⚫ OrderDate ✖OR⚓ a.LASTEDITEDWHEN is NULL➖) ❌where⚓ a.PURCHASEORDERID ⚫=⚫ 34559 ✖and⚓ a.STOCKITEMID ⚫=⚫ 1 ❌QUALIFY⚓ ROW_NUMBER(➕➖) OVER(➕ PARTITION⧆ BY ORDEREDOUTERS ❌ORDER⚓⧆ BY RECEIVEDOUTERS DESC➖) ⚫=⚫ 1⛓⚓➖)❌,purchase_con AS(➕❌ select⚓ DISTINCT CONCAT(➕pol.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,dl.OrderDate as ⛓⚓search_date✖⚓,pol.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sinv.ORDERID✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of the Voucher' as ⛓⚓ParentLevel_Desc ❌from⚓ PURCHASING_PURCHASEORDERLINES pol ❌inner⚓ ❌join⚓ SALES_ORDERS sord ❌on⚓ sord.ORDERID ⚫=⚫ pol.PURCHASEORDERID ✖and⚓ pol.PURCHASEORDERID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null ❌inner⚓ ❌join⚓ SALES_INVOICES sinv ❌on⚓ sinv.ORDERID ⚫=⚫ sord.ORDERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERS AS scus ❌ON⚓ scus.CUSTOMERID ⚫=⚫ sinv.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_CUSTOMERTRANSACTIONS AS sct ❌ON⚓ sct.CUSTOMERID ⚫=⚫ scus.CUSTOMERID ❌INNER⚓ ❌JOIN⚓ SALES_INVOICELINES AS sinvl ❌ON⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌INNER⚓ ❌join⚓ unique_purchase up ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌LEFT⚓ ❌join⚓ unique_item ui ❌on⚓ ui.STOCKITEMID ⚫=⚫ sinvl.STOCKITEMID ❌INNER⚓ ❌JOIN⚓ date_list dl ❌ON⚓ pol.LASTRECEIPTDATE⚫<=⚫ dl.OrderDate ✖AND⚓(➕ pol.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ pol.LASTEDITEDWHEN is NULL➖) ✖AND⚓ sinv.INVOICEDATE ⚫<=⚫ dl.OrderDate ✖AND⚓(➕ sinv.LASTEDITEDWHEN ⚫>⚫ dl.OrderDate ✖OR⚓ sinv.LASTEDITEDWHEN is NULL➖) ❌where⚓ pol.STOCKITEMID ⚫=⚫ 35 ✖and⚓ pol.STOCKITEMID ⚫=⚫ 1 ✖AND⚓ pol.ORDEREDOUTERS IS NOT NULL ✖and⚓ sord.SALESPERSONPERSONID ⚫=⚫ 1 ✖and⚓ sord.CONTACTPERSONID⚫=⚫ 1 ✖and⚓ sord.DELIVERYINSTRUCTIONS ⚫=⚫ 'SHR' ✖AND⚓ sord.CUSTOMERPURCHASEORDERNUMBER is not null⛓⚓➖)❌,rule AS(➕❌ select⚓ distinct CONCAT(➕orderedouters.ORDEREDOUTERS✖⚓,'-0'➖) as ⛓⚓ID✖⚓,orderedouters.OrderDate as ⛓⚓search_date✖⚓,orderedouters.ORDEREDOUTERS as ⛓⚓orderedouters✖⚓,sct.PAYMENTMETHODID✖⚓,sct.TRANSACTIONDATE✖⚓,up.PURCHASEORDERLINEID as ⛓⚓PolID✖⚓,ui.ORDEREDOUTERS as ⛓⚓outersid✖⚓,UPPER(➕up.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,sinvl.STOCKITEMID✖⚓,0 as ⛓⚓ParentLevel✖⚓,'Direct Issuer of Salesperson' as ⛓⚓ParentLevel_Desc ❌from⚓ unique_orderedouters orderedouters ❌inner⚓ ❌join⚓ SALES_CUSTOMERTRANSACTIONS sct ❌on⚓ orderedouters.PURCHASEORDERLINEID ⚫=⚫ sct.PURCHASEORDERLINEID ❌inner⚓ ❌join⚓ SALES_INVOICELINES sinvl ❌on⚓ sinvl.INVOICEID ⚫=⚫ sct.INVOICEID ❌inner⚓ ❌join⚓ unique_purchase up ❌on⚓ up.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ❌LEFT⚓ ❌join⚓ unique_item up ❌on⚓ ui.PURCHASEORDERLINEID ⚫=⚫ sinvl.PURCHASEORDERLINEID ✖and⚓ ui.OrderDate ⚫=⚫ orderedouters.OrderDate ❌where⚓ sct.ISFINALIZED in(➕'TRUE'➖)⛓⚓➖)❌,direct_map AS(➕❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ purchase_con ❌UNION⚓❌ select⚓ ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc ❌from⚓ rule ❌where⚓ ID not in(➕ select⚓ distinct ID ❌from⚓ purchase_con➖)⛓⚓➖)❌,parent_child AS(➕❌ select⚓ Distinct upar.OrderDate AS ⛓⚓search_date✖⚓,upar.STOCKITEMIDas ⛓⚓PolID✖⚓,UPPER(➕cls.ORDEREDOUTERS➖) as ⛓⚓outersname✖⚓,cln.ORDEREDOUTERS as ⛓⚓outersid✖⚓,upar.ORDEREDOUTERS as ⛓⚓parent_PolID✖⚓,UPPER(➕pls.ORDEREDOUTERS➖) as ⛓⚓parent_outersname✖⚓,pln.ORDEREDOUTERS as ⛓⚓parent_outersid ❌from⚓ unique_parent upar ❌left⚓ ❌join⚓ unique_purchase uprc ❌on⚓ uprc.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uprc.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_purchase uprcs ❌on⚓ pls.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uprcs.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uit ❌on⚓ cln.STOCKITEMID ⚫=⚫ upar.STOCKITEMID ✖and⚓ uit.OrderDate ⚫=⚫ upar.OrderDate ❌left⚓ ❌join⚓ unique_item uitem ❌on⚓ uitem.STOCKITEMID ⚫=⚫ upar.ORDEREDOUTERS ✖and⚓ uitem.OrderDate ⚫=⚫ upar.OrderDate⛓⚓➖)❌,navigate_hierarchy as(➕❌ SELECT⚓ search_date✖⚓,PolID✖⚓,outersid✖⚓,cast(➕outersname as varchar➖) as ⛓⚓outersname✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,cast(➕parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,1 as ⛓⚓ParentLevel✖⚓,concat(➕ cast(➕outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕parent_outersname as varchar➖)➖) as ⛓⚓outersnameline ❌from⚓ parent_child ❌UNION ALL⚓❌ SELECT⚓ c.search_date✖⚓,c.PolID✖⚓,c.outersid✖⚓,cast(➕c.outersname as varchar➖) as ⛓⚓outersname✖⚓,p.parent_PolID✖⚓,p.parent_outersid✖⚓,cast(➕p.parent_outersname as varchar➖) as ⛓⚓parent_outersname✖⚓,p.ParentLevel ⚫+⚫ 1✖⚓,cast(➕ CONCAT(➕ cast(➕c.outersname as varchar➖)✖⚓,' -> '✖⚓,cast(➕p.outersnameline as varchar➖)➖) AS varchar➖) as ⛓⚓outersnameline ❌from⚓ parent_child C ❌join⚓ navigate_hierarchy p ❌on⚓ p.PolID ⚫=⚫ c.parent_PolID ✖AND⚓ c.PolID ⚫<>⚫ c.parent_PolID ✖and⚓ p.ParentLevel ⚫<=⚫ 40 ✖AND⚓ p.search_date ⚫=⚫ c.search_date ❌where⚓ p.ParentLevel ⚫<=⚫ 40⛓⚓➖)❌,TEST_OUTPUT AS(➕❌ select⚓ DISTINCT ID✖⚓,search_date✖⚓,orderedouters✖⚓,PAYMENTMETHODID✖⚓,TRANSACTIONDATE✖⚓,PolID✖⚓,outersid✖⚓,outersname✖⚓,STOCKITEMID✖⚓,ParentLevel✖⚓,ParentLevel_Desc✖⚓,PolID as ⛓⚓parent_PolID✖⚓,outersid as ⛓⚓parent_outersid✖⚓,outersname as ⛓⚓parent_outersname✖⚓,outersname as ⛓⚓outersnameline ❌from⚓ direct_mapping ❌UNION⚓❌ select⚓ CONCAT(➕dm.orderedouters✖⚓,'-'✖⚓,nh.Parentlevel➖) as ⛓⚓ID✖⚓,dm.search_date✖⚓,dm.orderedouters✖⚓,dm.PAYMENTMETHODID✖⚓,dm.TRANSACTIONDATE✖⚓,nh.PolID✖⚓,nh.outersid✖⚓,nh.outersname✖⚓,dm.STOCKITEMID✖⚓,nh.ParentLevel✖⚓,❌CASE⚓➕ ❌WHEN⚓ nh.ParentLevel ⚫=⚫ 1 ❌THEN⚓ 'Immediate Parent of Salesperson Issuer' ❌ELSE⚓ concat(➕ 'Higher Level Parent ('✖⚓,nh.ParentLevel✖⚓,') Levels above the Salesperson Issuer'➖) ➖❌END⚓ as ⛓⚓ParentLevel_Desc✖⚓,parent_PolID✖⚓,parent_outersid✖⚓,parent_outersname✖⚓,outersnameline as ⛓⚓outersnameline ❌from⚓ navigate_hierarchy nh ❌inner⚓ ❌join⚓ direct_map dm ❌on⚓ dm.PolID ⚫=⚫ nh.PolID⛓⚓➖)❌SELECT⚓ *❌FROM⚓ TEST_OUTPUT ❌group⚓ by search_date"""

        expected = """WITH date_list AS(
    SELECT CAST('2021-05-24' AS date) AS OrderDate 
     UNION
    SELECT CAST('2021-05-25' AS date) AS OrderDate)
,unique_orderedouters AS(
    select PURCHASEORDERLINEID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE<=OrderDate
       AND( a.LASTEDITEDWHEN>OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID=27229
       and a.STOCKITEMID=1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC)=1)
,unique_purchase as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE<=OrderDate
       AND( a.LASTEDITEDWHEN>OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID=27033
       and a.STOCKITEMID=1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC)=1)
,unique_item as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE<=OrderDate
       AND( a.LASTEDITEDWHEN>OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID=114364
       and a.STOCKITEMID=1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC)=1)
,unique_parent as(
    select STOCKITEMID
          ,ORDEREDOUTERS
          ,dl.OrderDate 
      from PURCHASING_PURCHASEORDERLINES a 
      JOIN date_list dl 
        ON a.LASTRECEIPTDATE<=OrderDate
       AND( a.LASTEDITEDWHEN>OrderDate
            OR a.LASTEDITEDWHEN is NULL) 
     where a.PURCHASEORDERID=34559
       and a.STOCKITEMID=1
   QUALIFY ROW_NUMBER() OVER( PARTITION BY ORDEREDOUTERS 
         ORDER BY RECEIVEDOUTERS DESC)=1)
,purchase_con AS(
    select DISTINCT CONCAT(pol.ORDEREDOUTERS
              ,'-0') as ID
          ,dl.OrderDate as search_date
          ,pol.ORDEREDOUTERS as orderedouters
          ,sinv.ORDERID
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as PolID
          ,ui.ORDEREDOUTERS as outersid
          ,UPPER(up.ORDEREDOUTERS) as outersname
          ,sinvl.STOCKITEMID
          ,0 as ParentLevel
          ,'Direct Issuer of the Voucher' as ParentLevel_Desc 
      from PURCHASING_PURCHASEORDERLINES pol 
     inner 
      join SALES_ORDERS sord 
        on sord.ORDERID=pol.PURCHASEORDERID
       and pol.PURCHASEORDERID=35
       and pol.STOCKITEMID=1
       and sord.SALESPERSONPERSONID=1
       and sord.CONTACTPERSONID=1
       and sord.DELIVERYINSTRUCTIONS='SHR'
       AND pol.ORDEREDOUTERS IS NOT NULL 
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null 
     inner 
      join SALES_INVOICES sinv 
        on sinv.ORDERID=sord.ORDERID
     INNER 
      JOIN SALES_CUSTOMERS AS scus 
        ON scus.CUSTOMERID=sinv.CUSTOMERID
     INNER 
      JOIN SALES_CUSTOMERTRANSACTIONS AS sct 
        ON sct.CUSTOMERID=scus.CUSTOMERID
     INNER 
      JOIN SALES_INVOICELINES AS sinvl 
        ON sinvl.INVOICEID=sct.INVOICEID
     INNER 
      join unique_purchase up 
        on ui.STOCKITEMID=sinvl.STOCKITEMID
      LEFT 
      join unique_item ui 
        on ui.STOCKITEMID=sinvl.STOCKITEMID
     INNER 
      JOIN date_list dl 
        ON pol.LASTRECEIPTDATE<=dl.OrderDate
       AND( pol.LASTEDITEDWHEN>dl.OrderDate
            OR pol.LASTEDITEDWHEN is NULL) 
       AND sinv.INVOICEDATE<=dl.OrderDate
       AND( sinv.LASTEDITEDWHEN>dl.OrderDate
            OR sinv.LASTEDITEDWHEN is NULL) 
     where pol.STOCKITEMID=35
       and pol.STOCKITEMID=1
       AND pol.ORDEREDOUTERS IS NOT NULL 
       and sord.SALESPERSONPERSONID=1
       and sord.CONTACTPERSONID=1
       and sord.DELIVERYINSTRUCTIONS='SHR'
       AND sord.CUSTOMERPURCHASEORDERNUMBER is not null)
,rule AS(
    select distinct CONCAT(orderedouters.ORDEREDOUTERS
              ,'-0') as ID
          ,orderedouters.OrderDate as search_date
          ,orderedouters.ORDEREDOUTERS as orderedouters
          ,sct.PAYMENTMETHODID
          ,sct.TRANSACTIONDATE
          ,up.PURCHASEORDERLINEID as PolID
          ,ui.ORDEREDOUTERS as outersid
          ,UPPER(up.ORDEREDOUTERS) as outersname
          ,sinvl.STOCKITEMID
          ,0 as ParentLevel
          ,'Direct Issuer of Salesperson' as ParentLevel_Desc 
      from unique_orderedouters orderedouters 
     inner 
      join SALES_CUSTOMERTRANSACTIONS sct 
        on orderedouters.PURCHASEORDERLINEID=sct.PURCHASEORDERLINEID
     inner 
      join SALES_INVOICELINES sinvl 
        on sinvl.INVOICEID=sct.INVOICEID
     inner 
      join unique_purchase up 
        on up.PURCHASEORDERLINEID=sinvl.PURCHASEORDERLINEID
      LEFT 
      join unique_item up 
        on ui.PURCHASEORDERLINEID=sinvl.PURCHASEORDERLINEID
       and ui.OrderDate=orderedouters.OrderDate
     where sct.ISFINALIZED in('TRUE'))
,direct_map AS(
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from purchase_con 
     UNION
    select ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc 
      from rule 
     where ID not in( select distinct ID 
                         from purchase_con))
,parent_child AS(
    select Distinct upar.OrderDate AS search_date
          ,upar.STOCKITEMIDas PolID
          ,UPPER(cls.ORDEREDOUTERS) as outersname
          ,cln.ORDEREDOUTERS as outersid
          ,upar.ORDEREDOUTERS as parent_PolID
          ,UPPER(pls.ORDEREDOUTERS) as parent_outersname
          ,pln.ORDEREDOUTERS as parent_outersid 
      from unique_parent upar 
      left 
      join unique_purchase uprc 
        on uprc.STOCKITEMID=upar.STOCKITEMID
       and uprc.OrderDate=upar.OrderDate
      left 
      join unique_purchase uprcs 
        on pls.STOCKITEMID=upar.ORDEREDOUTERS
       and uprcs.OrderDate=upar.OrderDate
      left 
      join unique_item uit 
        on cln.STOCKITEMID=upar.STOCKITEMID
       and uit.OrderDate=upar.OrderDate
      left 
      join unique_item uitem 
        on uitem.STOCKITEMID=upar.ORDEREDOUTERS
       and uitem.OrderDate=upar.OrderDate)
,navigate_hierarchy as(
    SELECT search_date
          ,PolID
          ,outersid
          ,cast(outersname as varchar) as outersname
          ,parent_PolID
          ,parent_outersid
          ,cast(parent_outersname as varchar) as parent_outersname
          ,1 as ParentLevel
          ,concat( cast(outersname as varchar)
                             ,' -> '
                             ,cast(parent_outersname as varchar)) as outersnameline 
      from parent_child 
   UNION ALL
    SELECT c.search_date
          ,c.PolID
          ,c.outersid
          ,cast(c.outersname as varchar) as outersname
          ,p.parent_PolID
          ,p.parent_outersid
          ,cast(p.parent_outersname as varchar) as parent_outersname
          ,p.ParentLevel+1
          ,cast( CONCAT( cast(c.outersname as varchar)
                  ,' -> '
                  ,cast(p.outersnameline as varchar)) AS varchar) as outersnameline 
      from parent_child C 
      join navigate_hierarchy p 
        on p.PolID=c.parent_PolID
       AND c.PolID<>c.parent_PolID
       and p.ParentLevel<=40
       AND p.search_date=c.search_date
     where p.ParentLevel<=40)
,TEST_OUTPUT AS(
    select DISTINCT ID
          ,search_date
          ,orderedouters
          ,PAYMENTMETHODID
          ,TRANSACTIONDATE
          ,PolID
          ,outersid
          ,outersname
          ,STOCKITEMID
          ,ParentLevel
          ,ParentLevel_Desc
          ,PolID as parent_PolID
          ,outersid as parent_outersid
          ,outersname as parent_outersname
          ,outersname as outersnameline 
      from direct_mapping 
     UNION
    select CONCAT(dm.orderedouters
                             ,'-'
                             ,nh.Parentlevel) as ID
          ,dm.search_date
          ,dm.orderedouters
          ,dm.PAYMENTMETHODID
          ,dm.TRANSACTIONDATE
          ,nh.PolID
          ,nh.outersid
          ,nh.outersname
          ,dm.STOCKITEMID
          ,nh.ParentLevel
          ,
      CASE 
                         WHEN nh.ParentLevel=1
                         THEN 'Immediate Parent of Salesperson Issuer' 
                         ELSE concat( 'Higher Level Parent ('
                  ,nh.ParentLevel
                  ,') Levels above the Salesperson Issuer') 
       END as ParentLevel_Desc
          ,parent_PolID
          ,parent_outersid
          ,parent_outersname
          ,outersnameline as outersnameline 
      from navigate_hierarchy nh 
     inner 
      join direct_map dm 
        on dm.PolID=nh.PolID)
SELECT *
  FROM TEST_OUTPUT 
 group by search_date"""

        self.assertEqual(expected, _process_data(marked_sql, "compact"))




from flowhigh.model.TypeCast import TypeCast
from flowhigh.model.TreeNode import TreeNode


class AntiPattern(TypeCast, TreeNode):
    severity: str = None
    readability: str = None
    correctness: str = None
    performance: str = None
    pos: list = []
    link: str = None
    type_: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class AntiPatternBuilder (object):
    construction: AntiPattern

    
    ap_links: dict = {'AP_01': {'name': 'Avoid ANSI-89 join syntax', 'severity': 'Warning', 'performance': ' ', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-ansi-89-join-syntax/'}, 'AP_03': {'name': 'Avoid count(*) in the outer join', 'severity': 'Warning', 'performance': 'Y', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-count-in-the-outer-join/'}, 'AP_09': {'name': 'Avoid filtering attributes from the non-preserved side of an outer join', 'severity': 'Warning', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-filtering-attributes-from-the-non-preserved-side-of-an-outer-join/'}, 'AP_11': {'name': 'Avoid implicit column references', 'severity': 'Warning', 'performance': ' ', 'readability': 'Y', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-implicit-column-references/'}, 'AP_14': {'name': 'Avoid implicit cross join', 'severity': 'Warning', 'performance': ' ', 'readability': 'Y', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-implicit-cross-join/'}, 'AP_13': {'name': 'Avoid inner join after outer join in multi-join query', 'severity': 'Warning', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-inner-join-after-outer-join-in-multi-join-query/'}, 'AP_04': {'name': 'Avoid nesting scalar subqueries in the SELECT statement', 'severity': 'Warning', 'performance': 'Y', 'readability': 'Y', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-nesting-scalar-subqueries-in-the-select-statement/'}, 'AP_30': {'name': 'Avoid or delay ORDER BY in inline views', 'severity': 'Caution', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-or-delay-order-by-in-inline-views/'}, 'AP_18': {'name': 'Avoid ordinal numbers when using ORDER BY or GROUP BY', 'severity': 'Caution', 'performance': ' ', 'readability': 'Y', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-ordinal-numbers-when-using-order-by-or-group-by/'}, 'AP_23': {'name': 'Avoid the natural join clause', 'severity': 'Warning', 'performance': ' ', 'readability': 'Y', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-the-natural-join-clause/'}, 'AP_22': {'name': 'Avoid UNION for multiple filter values', 'severity': 'Caution', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-union-for-multiple-filter-values/'}, 'AP_20': {'name': 'Avoid unused Common Table Expressions (CTE)', 'severity': 'Warning', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-unused-common-table-expressions-cte/'}, 'AP_28': {'name': 'Avoid using functions in the Join clause', 'severity': 'Caution', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-using-functions-in-the-join-clause/'}, 'AP_07': {'name': 'Avoid using functions in the WHERE clause', 'severity': 'Caution', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-using-functions-in-the-where-clause/'}, 'AP_26': {'name': 'Avoid using WHERE to filter aggregate columns', 'severity': 'Warning', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/avoid-using-where-to-filter-aggregate-columns/'}, 'AP_02': {'name': 'Beware of count distinct', 'severity': 'Notice', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-count-distinct/'}, 'AP_10': {'name': 'Beware of filtering for NULL', 'severity': 'Warning', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-filtering-for-null/'}, 'AP_12': {'name': 'Beware of implicit self-joins in a correlated subquery', 'severity': 'Caution', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-implicit-self-joins-in-a-correlated-subquery/'}, 'AP_16': {'name': 'Beware of NULL in combination with not equal operator (!=, <>)', 'severity': 'Warning', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-null-in-combination-with-not-equal-operator/'}, 'AP_08': {'name': 'Beware of NULL with arithmetic and string operations', 'severity': 'Caution', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-null-with-arithmetic-and-string-operations/'}, 'AP_05': {'name': 'Beware of SELECT *', 'severity': 'Caution', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-select-star/'}, 'AP_06': {'name': 'Beware of SELECT DISTINCT', 'severity': 'Notice', 'performance': 'Y', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/beware-of-select-distinct/'}, 'AP_19': {'name': 'Split multi join queries into smaller chunks', 'severity': 'Caution', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/split-multi-join-queries-into-smaller-chunks/'}, 'AP_15': {'name': 'Use an alias for derived columns', 'severity': 'Warning', 'performance': ' ', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-an-alias-for-derived-columns/'}, 'AP_27': {'name': 'Use an alias for inline views', 'severity': 'Caution', 'performance': ' ', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-an-alias-for-inline-views/'}, 'AP_17': {'name': 'Use IN / NOT IN for multiple OR operators', 'severity': 'Caution', 'performance': ' ', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-in-or-not-in-for-multiple-or-operators/'}, 'AP_24': {'name': 'Use LIKE instead of REGEX where LIKE is possible', 'severity': 'Warning', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-like-instead-of-regex-where-like-is-possible/'}, 'AP_29': {'name': 'Use parentheses when mixing ANDs with ORs', 'severity': 'Caution', 'performance': ' ', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-parentheses-when-mixing-ands-with-ors/'}, 'AP_25': {'name': 'Use UNION ALL instead of UNION', 'severity': 'Caution', 'performance': 'Y', 'readability': ' ', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-union-all-instead-of-union/'}, 'AP_21': {'name': 'Use window functions instead of self joins', 'severity': 'Caution', 'performance': 'Y', 'readability': 'Y', 'correctness': ' ', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/use-window-functions-instead-of-self-joins/'}, 'AP_33': {'name': 'WHERE NOT IN without NOT NULL check in subquery', 'severity': 'Caution', 'performance': ' ', 'readability': ' ', 'correctness': 'Y', 'link': 'https://docs.sonra.io/flowhigh/dev/docs/sql-optimiser/where-not-in-without-not-null-check-in-subquery/'}}
    
    def __init__(self) -> None:
        super().__init__()
        self.construction = AntiPattern()
    
    def with_severity(self, severity: str):
        if isinstance(severity, str) and severity.startswith("AP_"):
            child = self.ap_links[severity]["severity"]
        else:
            child = None
        self.construction.severity = child
    
    def with_readability(self, readability: str):
        if isinstance(readability, str) and readability.startswith("AP_"):
            child = self.ap_links[readability]["readability"]
        else:
            child = None
        self.construction.readability = child
    
    def with_correctness(self, correctness: str):
        if isinstance(correctness, str) and correctness.startswith("AP_"):
            child = self.ap_links[correctness]["correctness"]
        else:
            child = None
        self.construction.correctness = child
    
    def with_performance(self, performance: str):
        if isinstance(performance, str) and performance.startswith("AP_"):
            child = self.ap_links[performance]["performance"]
        else:
            child = None
        self.construction.performance = child
    
    def with_pos(self, pos: list):
        child = pos
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), pos)):
            self.construction.add_child(node)
        self.construction.pos = child
    
    def with_link(self, link: str):
        if isinstance(link, str) and link.startswith("AP_"):
            child = self.ap_links[link]["link"]
        else:
            child = None
        self.construction.link = child
    
    def with_type(self, type_: str):
        child = type_
        self.construction.type_ = child

    def build(self):
        return self.construction




from flowhigh.model.Expr import Expr
from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class Position(BaseExpr):
    string: Expr = None
    subString: Expr = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class PositionBuilder (object):
    construction: Position

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Position()
    
    def with_string(self, string: Expr):
        child = string
        if TreeNode in Expr.mro():
            self.construction.add_child(child)
        self.construction.string = child
    
    def with_subString(self, subString: Expr):
        child = subString
        if TreeNode in Expr.mro():
            self.construction.add_child(child)
        self.construction.subString = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction



import weakref
from weakref import WeakSet


class RegistryBase(object):
    ENTITY_REGISTRY = WeakSet()

    def __new__(cls, *args, **kwargs):
        pass

    @classmethod
    def get_registry(cls):
        return cls.ENTITY_REGISTRY

    @classmethod
    def push_to_registry(cls, instance):
        cls.ENTITY_REGISTRY.add(instance)

    @classmethod
    def search(cls, node_id):
        return next(filter(lambda x: x.get_id() == node_id, cls.get_registry()), None)


class AutoIncrementID(object):
    id: int = 0

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'id'):
            cls = super(AutoIncrementID, cls).__new__(cls)
        cls.id = cls.id + 1
        return cls.id


class TreeNode(object):
    _id: int
    _parent: weakref = None
    _children: WeakSet

    def __init__(self):
        self._id = AutoIncrementID()
        self._children = WeakSet()
        RegistryBase.push_to_registry(self)

    def get_id(self):
        return self._id

    def get_children(self) -> WeakSet:
        return self._children

    def get_parent(self):
        return self._parent

    def add_child(self, child) -> None:
        self._children.add(child)
        child._parent = weakref.proxy(self)

    def accept(self):
        pass





from flowhigh.model.Direction import Direction
from flowhigh.model.ExprExprCollectionHolder import ExprExprCollectionHolder


class InlineTable(ExprExprCollectionHolder):

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class InlineTableBuilder (object):
    construction: InlineTable

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = InlineTable()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_exprs(self, exprs: list):
        child = exprs
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), exprs)):
            self.construction.add_child(node)
        self.construction.exprs = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction



from enum import Enum


class DsAction(Enum):
    source = 1
    target = 2
    delete = 3
    insert = 4
    update = 5
    insert_overwrite = 6
    insert_overwrite_all = 7
    insert_all = 8
    create_table = 9
    create_table_local = 10
    create_table_global = 11
    create_table_global_volatile = 12
    create_table_local_volatile = 13
    create_table_global_transient = 14
    create_table_local_transient = 15
    copy_table = 16




from enum import Enum


class OpType(Enum):
    DEFAULT = 1
    CONCAT = 2
    CONCAT_PIPE = 3
    PLUS = 4
    MINUS = 5
    MUL = 6
    DIV_LIT = 7
    EQ = 8
    NEQ = 9
    AND = 10
    OR = 11
    NOT = 12
    LIKE = 13
    RLIKE = 14
    REGEX = 15
    NOT_LIKE = 16
    IS = 17
    IS_DISTINCT = 18
    IS_TRUE = 19
    IS_FALSE = 20
    IS_NULL = 21
    IS_UNKNOWN = 22
    LIKE_ALL = 23
    LIKE_SOME = 24
    LIKE_ANY = 25
    IS_NOT_DISTINCT = 26
    IS_DISTINCT_FROM = 27
    IS_NOT_DISTINCT_FROM = 28
    IN = 29
    ANY = 30
    ALL = 31
    MOD = 32
    LT = 33
    GT = 34
    GTE = 35
    LTE = 36
    IS_NOT = 37
    IS_NOT_NULL = 38
    IS_NOT_TRUE = 39
    IS_NOT_FALSE = 40
    IS_NOT_UNKNOWN = 41
    PARENTHESIS = 42
    NOT_EXISTS = 43
    DIV = 44
    MULTI = 45
    BETWEEN = 46
    NEQJ = 47
    EXISTS = 48
    INTERSECT = 49
    UNION = 50
    EXCEPT = 51
    SETMINUS = 52
    UNION_ALL = 53
    UNION_DISTINCT = 54
    REGEXP = 55
    ASSIGN = 56
    MATCH = 57
    NOT_MATCH = 58
    COLLATE = 59
    NOT_LIKE_ALL = 60
    NOT_LIKE_ANY = 61
    NOT_LIKE_SOME = 62
    NOT_IN = 63
    NOT_RLIKE_ALL = 64
    NOT_RLIKE_ANY = 65
    NOT_RLIKE_SOME = 66
    NOT_REGEXP_ALL = 67
    NOT_REGEXP_ANY = 68
    NOT_REGEXP_SOME = 69
    NOT_REGEXP = 70
    NOT_RLIKE = 71
    RLIKE_ANY = 72
    RLIKE_ALL = 73
    RLIKE_SOME = 74
    REGEXP_ANY = 75
    REGEXP_SOME = 76
    REGEXP_ALL = 77
    EQ_ALL = 78
    EQ_ANY = 79
    NEQ_ALL = 80
    NEQ_ANY = 81
    NEQJ_ALL = 82
    NEQJ_ANY = 83
    LT_ALL = 84
    LT_ANY = 85
    LTE_ALL = 86
    LTE_ANY = 87
    GT_ALL = 88
    GT_ANY = 89
    GTE_ALL = 90
    GTE_ANY = 91
    NSEQ_ANY = 92
    NSEQ_ALL = 93
    NSEQ = 94
    CASE = 95
    AMPERSAND = 96
    SOME = 97





from flowhigh.model.Direction import Direction
from flowhigh.model.Expr import Expr
from flowhigh.model.ExprExprHolder import ExprExprHolder


class Then(ExprExprHolder):

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class ThenBuilder (object):
    construction: Then

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Then()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_expr(self, expr: Expr):
        child = expr
        if TreeNode in Expr.mro():
            self.construction.add_child(child)
        self.construction.expr = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr
from flowhigh.model.TypeCast import TypeCast


class Current(BaseExpr, TypeCast):
    type_: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class CurrentBuilder (object):
    construction: Current

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Current()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_type(self, type_: str):
        child = type_
        self.construction.type_ = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.PageType import PageType
from flowhigh.model.Const import Const
from flowhigh.model.CoordinateBlock import CoordinateBlock
from flowhigh.model.TypeCast import TypeCast


class Page(CoordinateBlock, TypeCast):
    type_: PageType = None
    value: Const = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class PageBuilder (object):
    construction: Page

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Page()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_type(self, type_: PageType):
        child = type_
        self.construction.type_ = child
    
    def with_value(self, value: Const):
        child = value
        if TreeNode in Const.mro():
            self.construction.add_child(child)
        self.construction.value = child

    def build(self):
        return self.construction




from flowhigh.model.WrappedExpr import WrappedExpr
from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class Cast(BaseExpr):
    dataType: str = None
    expr: WrappedExpr = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class CastBuilder (object):
    construction: Cast

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Cast()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_dataType(self, dataType: str):
        child = dataType
        self.construction.dataType = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_expr(self, expr: WrappedExpr):
        child = expr
        if TreeNode in WrappedExpr.mro():
            self.construction.add_child(child)
        self.construction.expr = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.ParSeQLStatus import ParSeQLStatus
from flowhigh.model.CoordinateBlock import CoordinateBlock


class ParSeQL(CoordinateBlock):
    statement: list = []
    error: list = []
    version: str = None
    status: ParSeQLStatus = None
    ts: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class ParSeQLBuilder (object):
    construction: ParSeQL

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = ParSeQL()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_statement(self, statement: list):
        child = statement
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), statement)):
            self.construction.add_child(node)
        self.construction.statement = child
    
    def with_error(self, error: list):
        child = error
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), error)):
            self.construction.add_child(node)
        self.construction.error = child
    
    def with_version(self, version: str):
        child = version
        self.construction.version = child
    
    def with_ts(self, ts: str):
        child = ts
        self.construction.ts = child
    
    def with_status(self, status: ParSeQLStatus):
        child = status
        self.construction.status = child

    def build(self):
        return self.construction




from flowhigh.model.Ds import Ds
from flowhigh.model.CoordinateBlock import CoordinateBlock


class Copy(CoordinateBlock):
    fromExp: list = []
    targetColumns: list = []
    pattern: str = None
    copyOptions: str = None
    ds: list = []
    selectElements: list = []
    into: Ds = None
    fromStage: Ds = None
    partition: list = []
    file: str = None
    fromQuery: Ds = None
    header: bool = None
    fileFormat: str = None
    validation: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class CopyBuilder (object):
    construction: Copy

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Copy()
    
    def with_fromExp(self, fromExp: list):
        child = fromExp
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), fromExp)):
            self.construction.add_child(node)
        self.construction.fromExp = child
    
    def with_targetColumns(self, targetColumns: list):
        child = targetColumns
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), targetColumns)):
            self.construction.add_child(node)
        self.construction.targetColumns = child
    
    def with_pattern(self, pattern: str):
        child = pattern
        self.construction.pattern = child
    
    def with_copyOptions(self, copyOptions: str):
        child = copyOptions
        self.construction.copyOptions = child
    
    def with_ds(self, ds: list):
        child = ds
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), ds)):
            self.construction.add_child(node)
        self.construction.ds = child
    
    def with_selectElements(self, selectElements: list):
        child = selectElements
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), selectElements)):
            self.construction.add_child(node)
        self.construction.selectElements = child
    
    def with_fromStage(self, fromStage: Ds):
        child = fromStage
        if TreeNode in Ds.mro():
            self.construction.add_child(child)
        self.construction.fromStage = child
    
    def with_into(self, into: Ds):
        child = into
        if TreeNode in Ds.mro():
            self.construction.add_child(child)
        self.construction.into = child
    
    def with_file(self, file: str):
        child = file
        self.construction.file = child
    
    def with_partition(self, partition: list):
        child = partition
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), partition)):
            self.construction.add_child(node)
        self.construction.partition = child
    
    def with_fromQuery(self, fromQuery: Ds):
        child = fromQuery
        if TreeNode in Ds.mro():
            self.construction.add_child(child)
        self.construction.fromQuery = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_header(self, header: bool):
        child = header
        self.construction.header = child
    
    def with_validation(self, validation: str):
        child = validation
        self.construction.validation = child
    
    def with_fileFormat(self, fileFormat: str):
        child = fileFormat
        self.construction.fileFormat = child

    def build(self):
        return self.construction




from flowhigh.model.Direction import Direction
from flowhigh.model.ExprExprCollectionHolder import ExprExprCollectionHolder
from flowhigh.model.ReferencableExpr import ReferencableExpr
from flowhigh.model.TreeNode import TreeNode


class Asterisk(ExprExprCollectionHolder, ReferencableExpr, TreeNode):
    refsch: str = None
    fullref: str = None
    refdb: str = None
    refds: str = None
    refatt: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class AsteriskBuilder (object):
    construction: Asterisk

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Asterisk()
    
    def with_refsch(self, refsch: str):
        child = refsch
        self.construction.refsch = child
    
    def with_fullref(self, fullref: str):
        child = fullref
        self.construction.fullref = child
    
    def with_refdb(self, refdb: str):
        child = refdb
        self.construction.refdb = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_refds(self, refds: str):
        child = refds
        self.construction.refds = child
    
    def with_refatt(self, refatt: str):
        child = refatt
        self.construction.refatt = child
    
    def with_exprs(self, exprs: list):
        child = exprs
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), exprs)):
            self.construction.add_child(node)
        self.construction.exprs = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction
